# bib_util.py 

import string
#================================================
def is_maj(car):

        if string.find(string.ascii_uppercase,car) == -1:
                return 0
        else:
                return 1
# ===================================
def litfic( fic ):
        f = open(fic,"r")
        leTexte = f.readlines()
        texteConcat = ""
        f.close()
        for i in range(len(leTexte)):
                ch = leTexte[i]
                ch = ch.upper()
                for j in range(len(ch)):
                        if not is_maj(ch[j]):
                                continue
                        texteConcat = texteConcat + ch[j]
        return texteConcat
#======================================

